
import typing

from .types import T, P

__all__ = ['ArgumentBundle']


class ArgumentBundle(typing.Generic[P]):
    def __init__(self, *args: P.args, **kwargs: P.kwargs):
        self.args = args
        self.kwargs = kwargs
    
    def call(self, function: typing.Callable[P, T]) -> T:
        return function(*self.args, **self.kwargs)
    
    def to_string(self, indent: typing.Optional[int] = None) -> str:
        """
        Returns a string representation of the arguments.

        :param indent: The number of spaces to indent each line
        :return: A string representation of the arguments
        """
        chunks = []

        for value in self.args:
            chunks.append(repr(value))
        
        for name, value in self.kwargs.items():
            chunks.append(f'{name}={repr(value)}')
        
        if indent is None:
            arguments = ', '.join(chunks)
            return f'({arguments})'
        else:
            indent_str = ' ' * indent
            sep = ',\n' + indent_str
            arguments = sep.join(chunks)
            return f'(\n{indent_str}{arguments},\n)'
    
    def __repr__(self) -> str:
        cls_name = type(self).__qualname__
        return f'{cls_name}{self}'
    
    def __str__(self) -> str:
        return self.to_string()




import collections.abc
import re
import typing
import typing_extensions

from .introspection import get_type_arguments, is_parameterized_generic, get_generic_base_class
from .subtype_check import is_subtype
from .type_compat import to_python
from ..parameter import Parameter
from ..signature import Signature
from .._utils import eval_or_discard
from ..types import Type_

__all__ = ['is_instance']


Type_Variable = typing.TypeVar('Type_Variable', bound=Type_)


def is_instance(obj: typing.Any, type_: Type_Variable) -> typing_extensions.TypeGuard[Type_Variable]:
    """
    Returns whether ``obj`` is an instance of ``type_``. Unlike the builtin
    ``isinstance``, this function supports generics.

    .. warning::
        This function is a work-in-progress and only supports a subset of all
        generic types. If it encounters a type it can't handle, it will throw a
        :exc:`NotImplementedError`.

    .. versionadded:: 1.5
    """
    # Find out if the type has type parameters
    if not is_parameterized_generic(type_):
        if type_ in TESTS:
            test = TESTS[type_]
            return test(obj)
        
        cls = type(type_)
        if cls is typing.TypeVar:
            return _test_typevar(obj, type_)

        return isinstance(obj, type_)
    
    # Extract the generic base type and verify if the object is an instance of
    # that
    base_type = get_generic_base_class(type_)
    base_type = to_python(base_type, strict=False)

    if base_type in GENERIC_BASE_TESTS:
        test = GENERIC_BASE_TESTS[base_type]
        result = test(obj)
    else:
        result = isinstance(obj, base_type)
    
    if not result:
        return False
    
    # Verify the subtypes
    if base_type not in SUBTYPE_TESTS:
        raise NotImplementedError
    
    subtypes = get_type_arguments(type_)

    test = SUBTYPE_TESTS[base_type]
    return test(obj, *subtypes)


def _test_typevar(obj, var):
    if var.__bound__ is not None and not isinstance(obj, var.__bound__):
        return False
    
    if var.__constraints__:
        if not any(is_instance(obj, typ) for typ in var.__constraints__):
            return False
    
    return True


def _test_dict_subtypes(obj, key_type, value_type):
    for key, value in obj.items():
        if not is_instance(key, key_type):
            return False
        
        if not is_instance(value, value_type):
            return False
    
    return True


def _test_tuple_subtypes(obj, *subtypes):
    if len(obj) != len(subtypes):
        return False
    
    return all(
        is_instance(element, typ)
        for element, typ in zip(obj, subtypes)
    )


def _test_type_subtypes(obj, typ):
    return issubclass(obj, typ)


def _test_iterable_subtypes(obj, item_type):
    return all(
        is_instance(item, item_type)
        for item in obj
    )


def _test_annotated_subtypes(obj, typ, *_):
    return is_instance(obj, typ)


def _test_callable_subtypes(obj, param_types, return_type):
    signature = Signature.from_callable(obj)

    if signature.return_annotation is not Signature.empty:
        if not is_subtype(signature.return_annotation, return_type):
            return False
    
    if param_types is ...:
        return all(
            param.kind != Parameter.KEYWORD_ONLY
            for param in signature.parameters.values()
        )
    
    parameters = signature.parameter_list
    i = 0
    for param_type in param_types:
        if i >= len(parameters):
            return False

        param = parameters[i]
        
        if param.kind >= Parameter.KEYWORD_ONLY:
            return False
        
        if param.annotation is not Parameter.empty:
            # Functions are contravariant in their parameters, so the sub- and supertype are swapped here
            if not is_subtype(param_type, param.annotation):
                return False
        
        if param.kind != Parameter.VAR_POSITIONAL:
            i += 1
    
    # If any parameters with no default value remain, it's not a match
    return all(
        param.is_optional
        for param in parameters[i:]
    )


def _test_literal_subtypes(obj, *options):
    return obj in options


def _test_optional_subtypes(obj, typ):
    return obj is None or is_instance(obj, typ)


def _test_union_subtypes(obj, *types):
    return any(
        is_instance(obj, typ)
        for typ in types
    )


def _test_regex_pattern_subtypes(pattern, subtype):
    return isinstance(pattern.pattern, subtype)


def _test_regex_match_subtypes(match, subtype):
    return isinstance(match.string, subtype)


def _return_true(_):
    return True


TESTS = {
    typing.Any: _return_true,
}


GENERIC_BASE_TESTS = eval_or_discard({
    'typing.Literal': _return_true,
    'typing.Optional': _return_true,
    'typing.Union': _return_true,
}, globals())


SUBTYPE_TESTS = eval_or_discard({
    'dict': _test_dict_subtypes,
    'frozenset': _test_iterable_subtypes,
    'list': _test_iterable_subtypes,
    'set': _test_iterable_subtypes,
    'tuple': _test_tuple_subtypes,
    'type': _test_type_subtypes,

    'collections.abc.Callable': _test_callable_subtypes,
    'collections.abc.Iterable': _test_iterable_subtypes,

    'typing.Callable': _test_callable_subtypes,
    'typing.Literal': _test_literal_subtypes,
    'typing.Optional': _test_optional_subtypes,
    'typing.Union': _test_union_subtypes,

    'typing_extensions.Annotated': _test_annotated_subtypes,

    're.Pattern': _test_regex_pattern_subtypes,
    're.Match': _test_regex_match_subtypes,
}, globals())




import pytest

import introspection

import types

try:
    GenericAlias = types.GenericAlias
except:
    GenericAlias = ()

def pytest_make_parametrize_id(config, val, argname):
    # Represent types more accurately than pytest does by default.
    if getattr(val, '__module__', None) == 'typing' or isinstance(val, GenericAlias):
        return repr(val)


for cls in introspection.errors.Error.__subclasses__():
    @pytest.register_exception_compare(cls)  # If you get an error here, install pytest-raisin
    def my_error_compare(exc_actual, exc_expected):
        if vars(exc_actual) != vars(exc_expected):
            raise AssertionError(f'{exc_actual!r} != {exc_expected!r}')




import traceback
import typing
import warnings

__all__ = [
    'print_exception',
    'warn_from_exception',
]


def print_exception(exception: BaseException) -> None:
    """
    Shorthand for ``traceback.print_exception(type(exc), exc,
    exc.__traceback__)``.
    
    :param exception: The exception to print
    """
    traceback.print_exception(type(exception), exception, exception.__traceback__)


def warn_from_exception(
    exception: BaseException,
    warning_type: typing.Type[Warning] = UserWarning,
    message: typing.Optional[str] = None,
) -> None:
    """
    Turns the given exception into a warning, forwarding all the relevant
    information to :func:`warnings.warn_explicit`.

    :param warning_type: The class to use for the warning.
    :param message: A message for the warning. Defaults to ``str(exception)``.
    """
    if message is None:
        message = str(exception)
    
    tb = exception.__traceback__
    frame = tb.tb_frame
    code = frame.f_code

    module_globals = frame.f_globals

    warnings.warn_explicit(
        message,
        warning_type,
        filename=code.co_filename,
        lineno=tb.tb_lineno,
        module=module_globals.get('__name__'),
        registry=module_globals.get('__warningregistry__'),
        module_globals=module_globals,
    )




import pytest

from introspection import ArgumentBundle


@pytest.mark.parametrize('func, args, kwargs, expected', [
    (int, ['12'], {}, 12),
    (bytes, ['hi'], {'encoding': 'utf8'}, b'hi'),
])
def test_call(func, args, kwargs, expected):
    bundle = ArgumentBundle(*args, **kwargs)

    assert bundle.call(func) == expected


@pytest.mark.parametrize('args, kwargs, expected', [
    ([12], {}, "(12)"),
    (['hi'], {'encoding': 'utf8'}, "('hi', encoding='utf8')"),
])
def test_str(args, kwargs, expected):
    bundle = ArgumentBundle(*args, **kwargs)

    assert str(bundle) == expected


@pytest.mark.parametrize('args, kwargs, expected', [
    ([12], {}, "ArgumentBundle(12)"),
    (['hi'], {'encoding': 'utf8'}, "ArgumentBundle('hi', encoding='utf8')"),
])
def test_repr(args, kwargs, expected):
    bundle = ArgumentBundle(*args, **kwargs)

    assert repr(bundle) == expected




import typing

from ..types import Type_, ParameterizedGeneric

__all__ = ['parameterize']


def parameterize(type_: Type_, args: typing.Iterable) -> ParameterizedGeneric:
    """
    .. versionadded: 1.6
    """
    args = tuple(args)

    if len(args) == 1:
        args = args[0]
    
    return type_[args]



import types
import typing
import typing_extensions

import sentinel


NONE = sentinel.create("NONE")


GenericAliases = []
for alias in ("types.GenericAlias", "typing._GenericAlias"):
    try:
        GenericAliases.append(eval(alias))
    except AttributeError:
        pass
GenericAliases = tuple(GenericAliases)


T = typing.TypeVar("T")
P = typing_extensions.ParamSpec("P")
Class = typing.TypeVar("Class", bound=type)

Function = types.FunctionType
ParameterizedGeneric = typing.Union[GenericAliases]
Type_ = typing.Union[type, ParameterizedGeneric]
TypeAnnotation = typing.Union[Type_, str, typing.ForwardRef, None]


class GenericType(typing.Protocol):
    def __class_getitem__(cls, item) -> ParameterizedGeneric:
        ...


class Slot(typing.Protocol[T]):
    def __get__(self, instance: T, owner: typing.Optional[typing.Type[T]]) -> object:
        ...

    def __set__(self, instance: T, value: object) -> None:
        ...

    def __delete__(self, instance: T) -> None:
        ...




import pytest

import inspect
from introspection import Signature, BoundArguments


def test_bound_args_from_bound_arguments():
    sig = inspect.signature(lambda a, b, *c: None)
    bound_args = sig.bind('A', 'B', 'C', 'D')

    bound_args = BoundArguments.from_bound_arguments(bound_args)

    assert bound_args.arguments == {'a': 'A', 'b': 'B', 'c': ('C', 'D')}


@pytest.mark.parametrize('func, args, kwargs, expected_args, expected_kwargs', [
    ((lambda a, b, c:0), [9, 8, 7], {}, [9, 8, 7], {}),
    ((lambda a, b, c:0), [9, 8], {}, [9, 8], {}),
    ((lambda *b, c:0), [9, 8], {'c': 7}, [9, 8], {'c': 7}),
    (range, [9, 8, 7], {}, [9, 8, 7], {}),
    ((lambda a, *b, c:0), [9], {'c': 7}, [9], {'c': 7}),
    ((lambda a, *b, **c:0), [9, 8, 7], {'foo': 'bar'}, [9, 8, 7], {'foo': 'bar'}),
])
def test_to_varargs_prefer_args(func, args, kwargs, expected_args, expected_kwargs):
    sig = Signature.from_callable(func)
    bound_args = sig.bind_partial(*args, **kwargs)

    args, kwargs = bound_args.to_varargs(prefer='args')

    assert args == expected_args
    assert kwargs == expected_kwargs


@pytest.mark.parametrize('func, args, kwargs, expected_args, expected_kwargs', [
    ((lambda a, b, c:0), [9, 8, 7], {}, [], {'a': 9, 'b': 8, 'c': 7}),
    ((lambda a, b, c:0), [9, 8], {}, [], {'a': 9, 'b': 8}),
    ((lambda *b, c:0), [9, 8], {'c': 7}, [9, 8], {'c': 7}),
    (range, [9, 8, 7], {}, [9, 8, 7], {}),
    ((lambda a, *b, c:0), [9], {'c': 7}, [], {'a': 9, 'c': 7}),
    ((lambda a, *b, **c:0), [9, 8, 7], {'foo': 'bar'}, [9, 8, 7], {'foo': 'bar'}),
])
def test_to_varargs_prefer_kwargs(func, args, kwargs, expected_args, expected_kwargs):
    sig = Signature.from_callable(func)
    bound_args = sig.bind_partial(*args, **kwargs)

    args, kwargs = bound_args.to_varargs(prefer='kwargs')

    assert args == expected_args
    assert kwargs == expected_kwargs


def test_to_varargs_invalid_prefer_arg():
    bound_args = Signature.from_callable(lambda x: x).bind(3)

    with pytest.raises(ValueError):
        bound_args.to_varargs(prefer='foobar')


def test_iter():
    bound_args = Signature.from_callable(lambda x, y, z: x).bind_partial(3, 4)

    assert tuple(bound_args) == ('x', 'y')


def test_len():
    bound_args = Signature.from_callable(lambda x, y: x).bind_partial(3)

    assert len(bound_args) == 1


def test_getitem():
    bound_args = Signature.from_callable(lambda x: x).bind(3)

    assert bound_args['x'] == 3


def test_getitem_keyerror():
    bound_args = Signature.from_callable(lambda x: x).bind(3)

    with pytest.raises(KeyError):
        bound_args['y']


def test_setitem():
    bound_args = Signature.from_callable(lambda x: x).bind(3)

    bound_args['x'] = 17
    assert bound_args['x'] == 17


@pytest.mark.parametrize('func, args, kwargs, expected', [
    (lambda x: 0, [], {}, {'x'}),
    (lambda x=3: 0, [], {}, set()),
    (lambda x: 0, [3], {}, set()),
])
def test_get_missing_parameter_names(func, args, kwargs, expected):
    bound_args = Signature.from_callable(func).bind_partial(*args, **kwargs)

    assert bound_args.get_missing_parameter_names() == expected


@pytest.mark.parametrize('func, args, kwargs, expected', [
    (lambda x=3: 0, [], {}, {'x'}),
    (lambda *args: 0, [], {}, set()),
    (lambda **kwargs: 0, [], {}, set()),
])
def test_get_missing_parameter_names_with_optionals(func, args, kwargs, expected):
    bound_args = Signature.from_callable(func).bind_partial(*args, **kwargs)

    result = bound_args.get_missing_parameter_names(include_optional_parameters=True)
    assert result == expected




import collections
import functools
import inspect
from typing import *

from .argument_bundle import ArgumentBundle
from .misc import rename
from .parameter import Parameter
from .signature import Signature
from .types import P, T, Type_, NONE


__all__ = [
    'signature', 'get_parameters',
    'set_signature', 'replace_varargs', 'wraps',
    'split_arguments',
]


def signature(*args, **kwargs) -> Signature:
    """
    Shorthand for :meth:`Signature.from_callable`.
    """
    return Signature.from_callable(*args, **kwargs)


def get_parameters(callable_: Callable) -> List[Parameter]:
    """
    Returns a list of parameters accepted by ``callable_``.

    :param callable_: The callable whose parameters to retrieve
    :return: A list of :class:`Parameter` instances
    """
    return list(Signature.from_callable(callable_).parameters.values())


@overload
def set_signature(
    signature: Union[Callable[P, T], Signature[P], inspect.Signature],
    remove_parameters: None = None,
) -> Callable[[Callable], Callable[P, T]]: ...

@overload
def set_signature(
    signature: Union[Callable[..., T], inspect.Signature],
    remove_parameters: Iterable[Union[str, int, inspect._ParameterKind]],
) -> Callable[[Callable], Callable[..., T]]: ...

def set_signature(
    signature: Union[Callable[P, T], inspect.Signature],
    remove_parameters: Optional[Iterable[Union[str, int, inspect._ParameterKind]]] = None,
) -> Callable[[Callable], Callable[P, T]]:
    if not isinstance(signature, inspect.Signature):
        signature = Signature.from_callable(signature)
    
    if remove_parameters:
        if not isinstance(signature, Signature):
            signature = Signature.from_signature(signature)

        signature = signature.without_parameters(*remove_parameters)
    
    def decorator(func: Callable) -> Callable[P, T]:
        func.__signature__ = signature
        return func  # type: ignore

    return decorator


def replace_varargs(
    signature: Union[Callable[..., T], inspect.Signature],
    remove_parameters: Optional[Iterable[Union[str, int, inspect._ParameterKind]]] = None,
) -> Callable[[Callable], Callable[..., T]]:
    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        merged_signature = Signature.from_callable(func).replace_varargs(signature)

        if remove_parameters:
            merged_signature = merged_signature.without_parameters(*remove_parameters)

        func.__signature__ = merged_signature
        return func

    return decorator


# Case 1: No parameters removed, return annotation unchanged
@overload
def wraps(
    wrapped_func: Callable[P, T],
    *,
    name: Optional[str] = None,
    signature: Union[None, inspect.Signature, Callable] = None,
    remove_parameters: None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...

# Case 2: No parameters removed, return annotation changed
@overload
def wraps(
    wrapped_func: Callable[P, Any],
    *,
    name: Optional[str] = None,
    signature: Union[None, inspect.Signature, Callable] = None,
    remove_parameters: None = None,
    return_annotation: Type[T],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...

# Case 3: Parameters removed, return annotation unchanged
@overload
def wraps(
    wrapped_func: Callable[..., T],
    *,
    name: Optional[str] = None,
    signature: Union[None, inspect.Signature, Callable] = None,
    remove_parameters: Iterable[Union[str, int, inspect._ParameterKind]],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...

# Case 4: Parameters removed, return annotation changed
@overload
def wraps(
    wrapped_func: Callable[..., Any],
    *,
    name: Optional[str] = None,
    signature: Union[None, inspect.Signature, Callable] = None,
    remove_parameters: Iterable[Union[str, int, inspect._ParameterKind]],
    return_annotation: Type[T],
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    ...

def wraps(
    wrapped_func: Callable,
    *,
    name: Optional[str] = None,
    signature: Union[None, inspect.Signature, Callable] = None,
    remove_parameters: Optional[Iterable[Union[str, int, inspect._ParameterKind]]] = None,
    return_annotation: Type_ = NONE,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    Similar to :func:`functools.wraps`, but allows you to modify the function's
    metadata.

    .. versionadded:: 1.4

    :param wrapped_func: The wrapped function
    :param name: A new name for the wrapper function
    :param signature: A new signature for the wrapper function
    :param remove_parameters: Parameter names or indices to remove from the
        wrapper function's signature
    """
    def wrapper(wrapper_func: Callable[P, T]) -> Callable[P, T]:
        functools.update_wrapper(wrapper_func, wrapped_func)

        if name is not None:
            rename(wrapper_func, name)
        
        if (
            signature is not None or
            remove_parameters is not None or
            return_annotation is not NONE
        ):
            if signature is None:
                sig = Signature.from_callable(wrapped_func)
            elif isinstance(signature, inspect.Signature):
                sig = Signature.from_signature(signature)
            else:
                sig = Signature.from_callable(signature)
            
            if remove_parameters:
                sig = sig.without_parameters(*remove_parameters)
            
            if return_annotation is not NONE:
                sig = sig.replace(return_annotation=return_annotation)
            
            wrapper_func.__signature__ = sig

        return wrapper_func

    return wrapper


def split_arguments(
    signature: Union[Callable[P, Any], Signature[P], inspect.Signature],
    *args: object,
    **kwargs: object,
) -> Tuple[ArgumentBundle[P], ArgumentBundle]:
    """
    Given a signature (or callable) and arguments, splits the arguments into
    two pairs: Those that match the given signature, and those that don't.

    This is useful in situations where you would need multiple ``*args, **kwargs``
    for different purposes, for example::

        class Child(Parent1, Parent2):
            def __init__(self, *args, **kwargs):
                parent1_args, parent2_args = split_arguments(Parent1, *args, **kwargs)
                
                Parent1.__init__(self, *parent1_args.args, **parent1_args.kwargs)
                Parent2.__init__(self, *parent2_args.args, **parent2_args.kwargs)
    """
    if not isinstance(signature, inspect.Signature):
        signature = Signature.from_callable(signature)
    
    func_args = []
    func_kwargs = {}

    params = collections.deque(signature.parameters.values())

    # Positional arguments
    args_q = collections.deque(args)
    while params:
        if not args_q:
            break

        param = params[0]
        if param.kind > inspect.Parameter.POSITIONAL_OR_KEYWORD:
            break

        func_args.append(args_q.popleft())

        if param.kind != inspect.Parameter.VAR_POSITIONAL:
            del params[0]
    
    # Keyword arguments
    for param in params:
        if param.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.VAR_POSITIONAL):
            continue

        if param.kind == inspect.Parameter.VAR_KEYWORD:
            func_kwargs.update(kwargs)
            kwargs.clear()
            break
        
        try:
            value = kwargs.pop(param.name)
        except KeyError:
            continue

        func_kwargs[param.name] = value
    
    bundle1 = ArgumentBundle(*func_args, **func_kwargs)
    bundle2 = ArgumentBundle(*args_q, **kwargs)
    return bundle1, bundle2
